import uuid
from typing import List, Optional, Tuple

import click

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import auth_policy_create_and_update_params, globus_command
from gcs_cli.prettyprint import FORMAT_TEXT_RECORD, formatted_print
from gcs_cli.sdk_client.auth_client import MissingConsent
from gcs_cli.sdk_client.data import AuthPolicyDocument
from gcs_cli.services.auth import get_client


@globus_command(
    "create",
    short_help="Create an authentication policy",
    disable_options=["explicit_host", "format"],
)
@requires_login
@click.option(
    "--project-id",
    metavar="PROJECT_ID",
    type=click.UUID,
    help="Auth project ID where the new authentication policy will be stored",
)
@click.option(
    "--high-assurance",
    is_flag=True,
    default=False,
    help=(
        "Flag indicating if authentication policy will be used on high assurance "
        "collections. This setting is immutable."
    ),
)
@click.argument("DESCRIPTION", type=click.STRING)
@click.argument("DISPLAY_NAME", type=click.STRING)
@auth_policy_create_and_update_params(create=True)
def auth_policy_create(
    project_id: Optional[uuid.UUID],
    high_assurance: bool,
    include: Tuple[str],
    exclude: Tuple[str],
    description: str,
    display_name: str,
) -> None:
    client = get_client()

    # If the caller did not specify a project ID, we are going to lookup projects
    # owned by the caller and, if there is only a single project, we will use it.
    if project_id is None:
        try:
            project_res = client.get_project_list()
        except MissingConsent as e:
            raise click.ClickException(
                "Missing consent for the policy API. Please run the following "
                "command:\n"
                f"    globus-connect-server session consent {e.scope}"
            )

        if len(project_res["projects"]) == 0:
            raise click.ClickException(
                "You are not the admin of any projects. Please go to "
                "https://auth.globus.org/developers and create one."
            )
        if len(project_res["projects"]) > 1:
            raise click.ClickException(
                "You are the admin of multiple projects, please use the --project-id "
                "option to specify which project to use. Current projects:\n"
                + "\n".join(
                    [
                        f"{d['id']} : {d['project_name']}"
                        for d in project_res["projects"]
                    ]
                )
            )
        project_id = uuid.UUID(f"{project_res['projects'][0]['id']}")

    def _filter_domain_tuple(domain_tuple: Tuple[str, ...]) -> Optional[List[str]]:
        # Remove any duplicates
        domain_list = list(set(domain_tuple))
        # Remove any superfluous ''
        domain_list = list(filter(lambda x: x != "", domain_list))
        # Handle no values given
        if domain_list == []:
            return None
        return domain_list

    auth_policy_doc = AuthPolicyDocument(
        project_id=project_id,
        display_name=display_name,
        description=description,
        high_assurance=high_assurance,
        domain_constraints_include=_filter_domain_tuple(include),
        domain_constraints_exclude=_filter_domain_tuple(exclude),
    )

    try:
        policy_res = client.create_auth_policy(auth_policy_doc)
    except MissingConsent as e:
        raise click.ClickException(
            "Missing consent for the policy API. Please run the following "
            "command:\n"
            f"    globus-connect-server session consent {e.scope}"
        )

    formatted_print(
        policy_res,
        fields=[("Authentication Policy ID", "id")],
        text_format=FORMAT_TEXT_RECORD,
    )
