import uuid
from typing import Any, Optional, cast

import click

from gcs_cli.helpers import (
    load_certificate,
    load_certificate_chain,
    load_private_key,
    record_print_result_message,
    requires_login,
)
from gcs_cli.parsing import collection_id_arg, domain_update_options, globus_command
from gcs_cli.services.gcs import get_client


@globus_command(
    "update",
    short_help="Update the custom domain and certificates for a collection",
    disable_options=["format"],
)
@domain_update_options(endpoint=True)
@collection_id_arg
@requires_login
def domain_update(
    collection_id: uuid.UUID,
    certificate_path: str,
    certificate_chain_path: Optional[str],
    private_key_path: str,
    managed: bool,
    wildcard: bool,
    domain: str,
) -> None:
    """
    Update the domain to use for a collection. This can set either a wildcard
    or a non-wildcard domain. If a wildcard domain is used and the collection is
    a mapped collection, then all guest collections associated with this
    collection that don't have custom domains will be made subdomains of this
    domain. If not, only this collection's domain is changed.

    The certificate, chain, and key can either be managed or unmanaged. If
    managed, the data is encrypted and shared between data transfer nodes
    serving this endpoint. If unmanaged, the paths specified on the
    command-line must exist on each data transfer node, and Globus Connect
    Server will only share the location for those files between data transfer
    nodes.
    """
    client = get_client()

    doc = {
        "DATA_TYPE": "domain#1.0.0",
        "wildcard": wildcard,
        "domain_name": domain,
    }
    if not managed:
        doc["certificate_path"] = certificate_path
        if certificate_chain_path:
            doc["certificate_chain_path"] = certificate_chain_path
        doc["private_key_path"] = private_key_path
    else:
        doc["certificate"] = load_certificate(certificate_path)
        if certificate_chain_path:
            doc["certificate_chain"] = load_certificate_chain(certificate_chain_path)

        def get_passphrase(*args: Any, **kwargs: Any) -> bytes:
            return cast(str, click.prompt("Enter Passphrase", hide_input=True)).encode()

        doc["private_key"] = load_private_key(
            private_key_path, passphrase=get_passphrase
        )
    res = client.put_collection_domain(collection_id, doc)
    record_print_result_message(res)
