import uuid
from typing import Optional, Sequence, cast

import click

from globus_sdk.paging import Paginator

from gcs_cli.helpers import default_identity_id_resolver, requires_login
from gcs_cli.helpers.field_funcs import FieldSpec
from gcs_cli.parsing import globus_command, include_private_policies_opt
from gcs_cli.prettyprint import FORMAT_TEXT_TABLE, formatted_print
from gcs_cli.services.gcs import get_client

STANDARD_FIELDS: FieldSpec = [
    ("ID", "id"),
    ("Display Name", "display_name"),
    ("Owner", default_identity_id_resolver.field),
    ("Collection Type", "collection_type"),
    ("Storage Gateway ID", "storage_gateway_id"),
]


def hyphen_to_underscore_callback(
    ctx: click.Context, param: click.Parameter, value: Optional[Sequence[str]]
) -> Optional[Sequence[str]]:
    if value is None or len(value) == 0:
        return None
    else:
        return [x.replace("-", "_") for x in value]


class ChoiceAllowingUnderscoresAsHyphens(click.Choice):
    """Let people use "mapped-collections" or "mapped_collections" as a nicety"""

    def convert(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> Optional[str]:
        if value is None:
            return None
        value = value.replace("_", "-")
        return cast(str, super().convert(value, param, ctx))


@globus_command("list", short_help="List Collections")
@include_private_policies_opt
@requires_login
@click.option(
    "--mapped-collection-id",
    default=None,
    type=click.UUID,
    help=(
        "Filter results to Guest Collections on a specific Mapped Collection. "
        "This is the ID of the Mapped Collection"
    ),
)
@click.option(
    "--filter",
    "filters",
    multiple=True,
    # choices are shown with hyphens, but the command will receive them with underscores
    type=ChoiceAllowingUnderscoresAsHyphens(
        ["mapped-collections", "guest-collections", "managed-by-me", "created-by-me"]
    ),
    callback=hyphen_to_underscore_callback,
    help="""\
Filter results to one of the specified categories of collections. Can be applied
multiple times. Note that mutually exclusive filters are allowed and will find no
results.

The filters are as follows

\b
mapped-collections:
  Only show collections with collection_type="mapped"

\b
guest-collections:
  Only show collections with collection_type="guest"

\b
managed-by-me:
  Only show collections where one of your identities has a role

\b
created-by-me:
  Only show collections where one of your identities was the creator
""",
)
def collection_list(
    include_private_policies: bool,
    mapped_collection_id: uuid.UUID,
    filters: Sequence[str],
) -> None:
    """
    List Collections on the current Endpoint
    """
    client = get_client()
    params = {}
    if mapped_collection_id:
        params["mapped_collection_id"] = str(mapped_collection_id)
    if filters:
        params["filter"] = ",".join(filters)
    if include_private_policies:
        params["include"] = "private_policies"

    res: Paginator = client.paginated.get_collection_list(**params)

    formatted_print(res, text_format=FORMAT_TEXT_TABLE, fields=STANDARD_FIELDS)
