import logging
import uuid
from typing import Any, Dict, Union

from globus_sdk import GlobusHTTPResponse
from globus_sdk.response import IterableResponse
from globus_sdk.services.auth import AuthClient as SDKAuthClient

log = logging.getLogger(__name__)


# Todo: These response definitions exist here because the SDK doesn't yet
# support projects or policies. Once they do, these can be removed.


class GetPoliciesResponse(IterableResponse):
    default_iter_key = "policies"


class GetPolicyResponse(GlobusHTTPResponse):
    @property
    def data(self) -> Any:
        return self._parsed_json["policy"]


class GetProjectsResponse(IterableResponse):
    default_iter_key = "projects"


class MissingConsent(Exception):
    def __init__(self, scope: str, cause: Exception):
        self.scope = scope
        self.cause = cause

    def __str__(self) -> str:
        return str(self.cause)


class AuthClient(SDKAuthClient):
    def get_auth_policy_list(self, **params: Any) -> GetPoliciesResponse:
        """
        GET /v2/api/policies
        """
        path = "/v2/api/policies"
        return GetPoliciesResponse(self.get(path, query_params=params))

    def get_auth_policy(
        self, auth_policy_id: Union[uuid.UUID, str], **params: Any
    ) -> GetPolicyResponse:
        """
        GET /v2/api/policies/{auth_policy_id}
        """
        path = f"/v2/api/policies/{auth_policy_id}"
        return GetPolicyResponse(self.get(path, query_params=params))

    def create_auth_policy(
        self, data: Dict[str, Any], **params: Any
    ) -> GetPolicyResponse:
        """
        POST /v2/api/policies

        ** Parameters **

        ``data`` (*dict*)
        Data in the format of a Auth Policy document.
        """
        path = "/v2/api/policies"
        return GetPolicyResponse(
            self.post(path, data=data, query_params=params),
        )

    def update_auth_policy(
        self, auth_policy_id: Union[uuid.UUID, str], data: Dict[str, Any], **params: Any
    ) -> GetPolicyResponse:
        """
        PUT /v2/api/policies/{auth_policy_id}

        ** Parameters **

        ``data`` (*dict*)
        Data in the format of a Auth Policy document.
        """
        path = f"/v2/api/policies/{auth_policy_id}"
        return GetPolicyResponse(
            self.put(path, data=data, query_params=params),
        )

    def delete_auth_policy(
        self, auth_policy_id: Union[uuid.UUID, str], **params: Any
    ) -> GetPolicyResponse:
        """
        DELETE /v2/api/policies/{auth_policy_id}

        ** Parameters **

        """
        path = f"/v2/api/policies/{auth_policy_id}"
        return GetPolicyResponse(
            self.delete(path, query_params=params),
        )

    def get_project_list(self, **params: Any) -> GetProjectsResponse:
        """
        GET /v2/api/projects
        """
        path = "/v2/api/projects"
        return GetProjectsResponse(self.get(path, query_params=params))

    def request(self, *args: Any, **kwargs: Any) -> GlobusHTTPResponse:
        projects_scope = "urn:globus:auth:scope:auth.globus.org:manage_projects"
        try:
            return super().request(*args, **kwargs)
        except self.error_class as e:
            if e.http_status == 403 and projects_scope in e.message:  # noqa: B306
                raise MissingConsent(projects_scope, e)
            raise
