import os
import pathlib
import stat
from typing import cast, Generic, Optional, TypeVar, Union
from typing_extensions import TypedDict
import uuid

from ..typing.gcs_documents import (
    BLACKPEARL_CONNECTOR_ID,
    BlackPearlUserCredentialDict,
    BlackPearlUserCredentialPoliciesDict,
    CEPH_CONNECTOR_ID,
    CephUserCredentialDict,
    CephUserCredentialPoliciesDict,
    HPSS_CONNECTOR_ID,
    HPSSUserCredentialDict,
    HPSSUserCredentialPoliciesDict,
    POSIX_CONNECTOR_ID,
    PosixUserCredentialDict,
    PosixUserCredentialPoliciesDict,
    S3_CONNECTOR_ID,
    S3UserCredentialDict,
    S3UserCredentialPoliciesDict,
    UserCredentialDict,
)
from ..util import yaml, yaml_object


class CredentialYamlDict(TypedDict):
    username: Optional[str]
    subject: Optional[str]
    gsi_authz_callout_index: Optional[int]
    identity_id: Optional[str]
    identity_username: Optional[str]


class BlackPearlCredentialYamlDict(CredentialYamlDict):
    access_id: Optional[str]
    secret_key: Optional[str]


class HPSSCredentialYamlDict(CredentialYamlDict):
    pass


class PosixCredentialYamlDict(CredentialYamlDict):
    pass


class S3CredentialYamlDict(CredentialYamlDict):
    s3_key_id: Optional[str]
    s3_secret_key: Optional[str]


class CephCredentialYamlDict(CredentialYamlDict):
    pass


T = TypeVar("T", bound=CredentialYamlDict)
U = TypeVar("U", bound=UserCredentialDict)


@yaml_object(yaml)
class Credential(Generic[T, U]):
    namespace = uuid.UUID("8950b9aa-bffc-4353-a986-0ee3eced73da")

    def __init__(self, data: Optional[T] = None) -> None:
        self._id: Optional[str] = None
        self.username: Optional[str] = None
        self.subject: Optional[str] = None
        self.gsi_authz_callout_index: Optional[int] = None
        self.identity_id: Optional[str] = None
        self.identity_username: Optional[str] = None
        self.collection_id: Optional[str] = None
        self.storage_gateway_id: Optional[str] = None

        if data:
            self.__setstate__(data)

    @property
    def id(self) -> Optional[str]:
        if self.username and self.subject and not self._id:
            self._id = str(
                uuid.uuid5(Credential.namespace, f"{self.username}:{self.subject}")
            )
            return self._id
        return self._id

    @id.setter
    def id(self, value: Optional[str]) -> None:
        self._id = value

    def refresh_id(self) -> None:
        self._id = None

    def dump(self, root: Union[str, os.PathLike], omit_secrets: bool = False) -> None:
        path = pathlib.Path(root) / f"{self.id}.yaml"
        with open(
            os.open(
                path,
                os.O_CREAT | os.O_TRUNC | os.O_WRONLY,
                mode=stat.S_IWUSR | stat.S_IRUSR,
            ),
            "w",
        ) as f:
            yaml.dump(self, f)

    @staticmethod
    def load(
        root: Union[str, os.PathLike], filename: str, load_version: str
    ) -> Optional["Credential"]:
        path = pathlib.Path(root) / filename
        with open(path, "r") as f:
            obj = yaml.load(f)
            if isinstance(obj, PreConnectorCredential):
                obj = PosixCredential(obj.__getstate__())
            assert isinstance(obj, Credential)

        return obj

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Credential):
            return NotImplemented

        return self is other or self.__getstate__() == other.__getstate__()

    def __getstate__(self) -> T:
        return cast(
            T,
            {
                "username": self.username,
                "subject": self.subject,
                "gsi_authz_callout_index": self.gsi_authz_callout_index,
                "identity_id": self.identity_id,
                "identity_username": self.identity_username,
            },
        )

    def __setstate__(self, state: T) -> None:
        base_state = cast(CredentialYamlDict, state)
        self.username = base_state["username"]
        self.subject = base_state["subject"]
        self.gsi_authz_callout_index = base_state["gsi_authz_callout_index"]
        self.identity_id = base_state["identity_id"]
        self.identity_username = base_state["identity_username"]
        if not hasattr(self, "_id"):
            self._id = None

    def to_json(self) -> UserCredentialDict:
        user_credential: UserCredentialDict = {
            "DATA_TYPE": "user_credential#1.0.0",
        }
        if self.identity_id is not None:
            user_credential["identity_id"] = self.identity_id

        if self.storage_gateway_id:
            user_credential["storage_gateway_id"] = self.storage_gateway_id
        if self.username is not None:
            user_credential["username"] = self.username

        return user_credential


@yaml_object(yaml)
class BlackPearlCredential(
    Credential[BlackPearlCredentialYamlDict, BlackPearlUserCredentialDict]
):
    def __init__(self, data: Optional[BlackPearlCredentialYamlDict] = None) -> None:
        super().__init__()
        self.access_id: Optional[str] = None
        self.secret_key: Optional[str] = None
        if data:
            self.__setstate__(data)

    def to_json(self) -> BlackPearlUserCredentialDict:
        policies: BlackPearlUserCredentialPoliciesDict = {
            "DATA_TYPE": "blackpearl_user_credential_policies#1.0.0"
        }
        if self.access_id:
            policies["access_id"] = self.access_id
        if self.secret_key:
            policies["secret_key"] = self.secret_key

        user_credential = cast(BlackPearlUserCredentialDict, super().to_json())
        user_credential["policies"] = policies
        user_credential["connector_id"] = BLACKPEARL_CONNECTOR_ID

        return user_credential

    def dump(self, root: Union[str, os.PathLike], omit_secrets: bool = False) -> None:
        if omit_secrets:
            newobj = BlackPearlCredential(data=self.__getstate__())
            newobj.secret_key = None
            newobj.dump(root=root, omit_secrets=False)
        else:
            super().dump(root=root, omit_secrets=False)

    def __getstate__(self) -> BlackPearlCredentialYamlDict:
        state = super().__getstate__()
        state["access_id"] = self.access_id
        state["secret_key"] = self.secret_key

        return state

    def __setstate__(self, state: BlackPearlCredentialYamlDict) -> None:
        super().__setstate__(state)
        self.access_id = state["access_id"]
        self.secret_key = state["secret_key"]

    def __str__(self) -> str:
        subj_str = f" {self.subject}" if self.subject else ""
        username_str = f"{self.username}" if self.username else ""

        return f"<BlackPearlCredential{subj_str}:{username_str}>"


@yaml_object(yaml)
class HPSSCredential(Credential[HPSSCredentialYamlDict, HPSSUserCredentialDict]):
    def __init__(self, data: Optional[HPSSCredentialYamlDict] = None) -> None:
        super().__init__()
        if data:
            self.__setstate__(data)

    def to_json(self) -> HPSSUserCredentialDict:
        policies: HPSSUserCredentialPoliciesDict = {
            "DATA_TYPE": "hpss_user_credential_policies#1.0.0"
        }
        user_credential = cast(HPSSUserCredentialDict, super().to_json())
        user_credential["policies"] = policies
        user_credential["connector_id"] = HPSS_CONNECTOR_ID

        return user_credential

    def __str__(self) -> str:
        subj_str = f" {self.subject}" if self.subject else ""
        username_str = f"{self.username}" if self.username else ""

        return f"<HPSSCredential{subj_str}:{username_str}>"


@yaml_object(yaml)
class PosixCredential(Credential[PosixCredentialYamlDict, PosixUserCredentialDict]):
    def __init__(self, data: Optional[PosixCredentialYamlDict] = None) -> None:
        super().__init__()
        if data:
            self.__setstate__(data)

    def to_json(self) -> PosixUserCredentialDict:
        policies: PosixUserCredentialPoliciesDict = {
            "DATA_TYPE": "posix_user_credential_policies#1.0.0"
        }
        user_credential = cast(PosixUserCredentialDict, super().to_json())
        user_credential["policies"] = policies
        user_credential["connector_id"] = POSIX_CONNECTOR_ID

        return user_credential

    def __str__(self) -> str:
        subj_str = f" {self.subject}" if self.subject else ""
        username_str = f"{self.username}" if self.username else ""

        return f"<PosixCredential{subj_str}:{username_str}>"


@yaml_object(yaml)
class S3Credential(Credential[S3CredentialYamlDict, S3UserCredentialDict]):
    def __init__(self, data: Optional[S3CredentialYamlDict] = None) -> None:
        super().__init__()
        self.s3_key_id: Optional[str] = None
        self.s3_secret_key: Optional[str] = None

        if data:
            self.__setstate__(data)

    def dump(self, root: Union[str, os.PathLike], omit_secrets: bool = False) -> None:
        if omit_secrets:
            newobj = S3Credential(data=self.__getstate__())
            newobj.s3_secret_key = None
            newobj.dump(root=root, omit_secrets=False)
        else:
            super().dump(root=root, omit_secrets=False)

    def __getstate__(self) -> S3CredentialYamlDict:
        state = super().__getstate__()
        state["s3_key_id"] = self.s3_key_id
        state["s3_secret_key"] = self.s3_secret_key

        return state

    def __setstate__(self, state: S3CredentialYamlDict) -> None:
        super().__setstate__(state)
        self.s3_key_id = state["s3_key_id"]
        self.s3_secret_key = state["s3_secret_key"]

    def to_json(self) -> S3UserCredentialDict:
        policies: S3UserCredentialPoliciesDict = {
            "DATA_TYPE": "s3_user_credential_policies#1.0.0",
        }
        if self.s3_key_id:
            policies["s3_key_id"] = self.s3_key_id
        if self.s3_secret_key:
            policies["s3_secret_key"] = self.s3_secret_key

        user_credential = cast(S3UserCredentialDict, super().to_json())
        user_credential["policies"] = policies
        user_credential["connector_id"] = S3_CONNECTOR_ID

        return user_credential

    def __str__(self) -> str:
        subj_str = f" {self.subject}" if self.subject else ""
        username_str = f"{self.username}" if self.username else ""

        return f"<S3Credential{subj_str}:{username_str}>"


@yaml_object(yaml)
class CephCredential(Credential[CephCredentialYamlDict, CephUserCredentialDict]):
    def __init__(self, data: Optional[CephCredentialYamlDict] = None) -> None:
        super().__init__()

        if data:
            self.__setstate__(data)

    def to_json(self) -> CephUserCredentialDict:
        policies: CephUserCredentialPoliciesDict = {
            "DATA_TYPE": "ceph_user_credential_policies#1.0.0",
        }

        user_credential = cast(CephUserCredentialDict, super().to_json())
        user_credential["policies"] = policies
        user_credential["connector_id"] = CEPH_CONNECTOR_ID

        return user_credential

    def __str__(self) -> str:
        subj_str = f" {self.subject}" if self.subject else ""
        username_str = f"{self.username}" if self.username else ""

        return f"<CephCredential{subj_str}:{username_str}>"


# Backward compatible storage gateway yaml object for deserializing
# credential.yaml from migration plans that predate multiple connector
# support
@yaml_object(yaml)
class PreConnectorCredential(PosixCredential):
    yaml_tag = "!Credential"
