from typing import Any, Dict, Optional, Tuple
from typing_extensions import TypedDict

from ..util import yaml, yaml_object
from ..parsers import GridFTPOptionParser
from ..typing.transfer_documents import EndpointServerDocumentDict


class NodeDict(TypedDict):
    ip_address: Optional[str]
    incoming_port_range: Optional[Tuple[int, int]]
    outgoing_port_range: Optional[Tuple[int, int]]
    data_interface: Optional[str]


@yaml_object(yaml)
class Node:
    def __init__(self) -> None:
        self.ip_address: Optional[str] = None
        self.incoming_port_range: Optional[Tuple[int, int]] = None
        self.outgoing_port_range: Optional[Tuple[int, int]] = None
        self.data_interface: Optional[str] = None
        self.transfer_endpoint_server: Optional[EndpointServerDocumentDict] = None
        self.gridftp_config: Optional[GridFTPOptionParser] = None

    @property
    def gridftp_config(self) -> Optional[GridFTPOptionParser]:
        return self._gridftp_config

    @gridftp_config.setter
    def gridftp_config(self, value: Optional[GridFTPOptionParser]) -> None:
        self._gridftp_config = value

        if value:
            if self.ip_address is None:
                self.ip_address = value.get_string_option("control_interface")
            if self.incoming_port_range is None:
                incoming_port_range = value.get_string_option("port_range")
                if incoming_port_range:
                    ports = [int(p) for p in incoming_port_range.split(",")]
                    assert len(ports) == 2
                    self.incoming_port_range = ports[0], ports[1]
            if self.outgoing_port_range is None:
                outgoing_port_range = value.get_string_option(
                    "$GLOBUS_TCP_SOURCE_RANGE"
                )
                if outgoing_port_range:
                    ports = [int(p) for p in outgoing_port_range.split(",")]
                    assert len(ports) == 2
                    self.outgoing_port_range = ports[0], ports[1]
            if self.data_interface is None:
                self.data_interface = value.get_string_option("data_interface")

    @staticmethod
    def from_transfer_endpoint_server(value: EndpointServerDocumentDict) -> "Node":
        node = Node()
        node.transfer_endpoint_server = value

        return node

    def to_json(self) -> Dict:
        ret: Dict[str, Any] = {
            "DATA_TYPE": "node#1.0.0",
        }
        if self.incoming_port_range:
            ret["incoming_port_range"] = list(self.incoming_port_range)
        if self.outgoing_port_range:
            ret["outgoing_port_range"] = list(self.outgoing_port_range)
        if self.data_interface:
            ret["data_interface"] = self.data_interface

        return ret

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Node):
            return NotImplemented

        return self is other or self.__getstate__() == other.__getstate__()

    def __getstate__(self) -> NodeDict:
        return {
            "ip_address": self.ip_address,
            "incoming_port_range": self.incoming_port_range,
            "outgoing_port_range": self.outgoing_port_range,
            "data_interface": self.data_interface,
        }

    def __setstate__(self, state: NodeDict) -> None:
        self.ip_address = state["ip_address"]
        self.incoming_port_range = state["incoming_port_range"]
        self.outgoing_port_range = state["outgoing_port_range"]
        self.data_interface = state["data_interface"]
