import itertools

from typing import Set
from .globus_client import get_api

from globus_sdk.response import GlobusHTTPResponse


def _get_identity_providers(domains: Set[str]) -> GlobusHTTPResponse:
    """
    Provide the Auth API method that is not available in the SDK as of
    version 3.4.0. This method will return the raw response.

    GET /v2/api/identity_providers?domains=globusid.org,google.com
    """
    auth_client = get_api("auth.globus.org")

    res: GlobusHTTPResponse = auth_client.get(
        "/v2/api/identity_providers", query_params={"domains": ",".join(domains)}
    )
    return res


def _get_usable_idp_domains(domains: Set[str]) -> Set[str]:
    """
    Returns the subset of 'domains' that are known to Auth and are not email
    identity providers.
    """
    res = _get_identity_providers(domains)

    # Get every domain listed for any returned email idp if the domain is in 'domains'
    return set(
        itertools.chain.from_iterable(
            (
                [d for d in idp["domains"] if d in domains]
                for idp in res.get("identity_providers", [])
                if idp["name"] != "Email Address"
            )
        )
    )


def get_unusable_idp_domains(domains: Set[str]) -> Set[str]:
    """
    Returns the subset of 'domains' that are either not recognized by Auth or
    are email identity providers.
    """
    return domains - _get_usable_idp_domains(domains)
