from typing import Dict, List, Optional
import uuid

from flask import (
    Response,
    current_app,
    g,
    jsonify,
    request,
    url_for,
)
from werkzeug.exceptions import HTTPException

from globus.manager.api.authentication import authenticated
from globus.manager.api.blueprints.decorators import (
    Base62Marker,
    expects_json,
    paginated,
    Paginator,
    query_parameter,
    requires_one_role,
    result_data,
)
from globus.manager.api.blueprints.utils import update_object
from globus.manager.api.models.result import Result
import globus.manager.api.errors as errors
import globus.manager.api.models as models
import globus.manager.api.schemas.api as api_schema
from .openapi import OpenAPIBlueprint


blueprint = OpenAPIBlueprint(
    "sharing_policies",
    __name__,
    tag="Sharing Policies",
    description="""
    The Sharing Restrictions Management API allows endpoint owners, endpoint
    administrators and collection administrators to manage sharing policies
    that contain information about both which users can share data and which
    paths in the connector's virtual filesystem they may share. These
    restrictions can only be created on a mapped collection, but are enforced
    on guest collection creation and access.
    """,
    menu_weight=10,
    url_prefix="/api/",
)


def get_collection_id_from_query() -> Optional[models.Collection]:
    collection_id = request.args.get("collection_id")
    collection = None
    collection = models.Collection.query.filter(
        ~models.Collection.is_deleted,  # type: ignore
        models.Collection.id == str(collection_id),
    ).one_or_none()

    g.collection = collection
    return collection


def get_collection_id_from_policy_doc() -> Optional[models.Collection]:
    try:
        json = request.get_json()
    except HTTPException:
        json = None

    if not json:
        raise errors.BadRequestError("Expected json input body.")

    collection_id = json.get("collection_id")
    collection = None

    if collection_id:
        collection = models.Collection.query.filter(
            ~models.Collection.is_deleted,  # type: ignore
            models.Collection.id == str(collection_id),
        ).one_or_none()

    g.collection = collection
    return collection


def get_collection_id_from_sharing_policy() -> Optional[models.Collection]:
    sharing_policy = None
    collection = None
    sharing_policy_id = str(request.view_args["sharing_policy_id"])
    sharing_policy = models.SharingPolicy.query.filter(
        ~models.SharingPolicy.is_deleted,  # type: ignore
        models.SharingPolicy.id == sharing_policy_id,
    ).one_or_none()

    if sharing_policy is not None:
        collection = sharing_policy.collection

    if sharing_policy is None:
        current_app.logger.debug("Couldn't find sharing policy %s", sharing_policy_id)
        raise errors.NotFoundError()

    g.collection = collection
    g.sharing_policy = sharing_policy

    return collection


@blueprint.route(
    "/sharing_policies",
    methods=["GET"],
    summary="List sharing policies",
    operationId="listSharingPolicies",
)
@result_data(
    description="List sharing policies response",
    schemas=[api_schema.sharing_policy_schema],
)
@authenticated
@query_parameter(
    name_spec="<uuid:collection_id>", description="ID of the collection", required=True
)
@query_parameter(
    name_spec="username", description="Username to query information about"
)
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_id_from_query,
)
@paginated
def sharing_policies_get(
    *,
    collection_id: uuid.UUID,
    username: Optional[str],
    page_size: int,
    marker: Optional[Paginator]
) -> Response:
    """
    List the sharing policies for a mapped collection.  This may return a
    paginated result; the `marker` and `page_size` query parameters can be used
    to obtain the next page of response data for the query.

    If the `username` query parameter is passed to this function then only the
    policies which are relevant to the given username are returned. This will
    include policies where the **users** property is `null` or contains the
    given username.
    """
    collection = g.collection
    if not collection or collection.is_guest_collection:
        raise errors.NotFoundError()

    next_page_token = None

    query = models.SharingPolicy.query.filter(
        models.SharingPolicy.collection_id == collection.id
    )

    if marker:
        query = query.filter(models.ConfigurationObject.id >= marker.marker)

    query = query.order_by(models.ConfigurationObject.id)

    results = []  # type: List[Dict]
    for sp in query:
        if sp.users and username and username not in sp.users:
            continue

        if len(results) < page_size:
            results.append(api_schema.sharing_policy_schema.dump(sp))
        else:
            next_page_token = Base62Marker(marker_data=sp.id)
            break

    res = api_schema.result_schema.dump(Result(data=results, marker=next_page_token))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route(
    "/sharing_policies",
    methods=["POST"],
    summary="Create a sharing policy",
    operationId="postSharingPolicy",
)
@result_data(
    schemas=[api_schema.sharing_policy_schema],
    description="List sharing policies response",
    max_items=1,
    http_status=201,
)
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_id_from_policy_doc,
)
@expects_json(
    schema=api_schema.sharing_policy_schema,
    description="New sharing policy",
    required=True,
)
def sharing_policies_post() -> Response:
    """
    Create a new sharing policy document for a mapped collection.  This new
    document will be added to the set of sharing policy documents for this
    collection. The sharing policy documents which either have no user
    constraint, or match the guest collection owner are used to determine which
    paths are available to be shared.

    This returns the policy document with the "id" field populated with the
    assigned ID of this policy.
    """
    collection = g.collection
    if not collection or collection.is_guest_collection:
        raise errors.NotFoundError()
    policy_doc = g.validated_data  # type: api_schema.SharingPolicyLoadDict

    assert policy_doc is not None

    if "id" in policy_doc:
        policy_doc.pop("id")

    if not policy_doc.get("users"):
        raise errors.ValidationError("Missing users value in input document")

    sharing_policy = None  # type: Optional[models.SharingPolicy]

    def update_collection(c):
        nonlocal sharing_policy
        sharing_policy = collection.create_sharing_policy(policy_doc)

    update_object(collection, update_collection)
    assert sharing_policy is not None

    g.updated_objects.append(str(collection.id))
    g.created_objects.append(str(sharing_policy.id))

    res = api_schema.result_schema.dump(
        Result(
            http_response_code=201,
            message="Created sharing policy {}".format(sharing_policy.id),
            data=[api_schema.sharing_policy_schema.dump(sharing_policy)],
        )
    )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    response.headers["Location"] = url_for(
        ".sharing_policies_get_by_id",
        sharing_policy_id=sharing_policy.id,
        _external=True,
        _scheme="https",
    )

    return response


@blueprint.route(
    "/sharing_policies/<uuid:sharing_policy_id>",
    methods=["GET"],
    summary="Get a sharing policy",
    path_parameters={"sharing_policy_id": "Id of the sharing policy"},
    operationId="getSharingPolicy",
)
@result_data(
    schemas=[api_schema.sharing_policy_schema],
    description="Get sharing policy response",
    max_items=1,
)
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_id_from_sharing_policy,
)
def sharing_policies_get_by_id(sharing_policy_id: uuid.UUID) -> Response:
    """
    Get a sharing policy.
    """
    collection = g.collection

    assert collection is not None and collection.is_mapped_collection

    sharing_policy = g.sharing_policy

    assert sharing_policy is not None

    res = api_schema.result_schema.dump(
        Result(data=[api_schema.sharing_policy_schema.dump(sharing_policy)])
    )

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route(
    "/sharing_policies/<uuid:sharing_policy_id>",
    methods=["DELETE"],
    summary="Delete a sharing policy",
    path_parameters={"sharing_policy_id": "Id of the sharing policy"},
    operationId="deleteSharingPolicy",
)
@result_data(
    schemas=[],
    description="Delete sharing policy response",
)
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_id_from_sharing_policy,
)
def sharing_policies_delete_by_id(sharing_policy_id: uuid.UUID) -> Response:
    """
    Delete a sharing policy.

    This may alter the behavior of existing guest collections, if the policies
    change the visible parts of the storage gateway's virtual file system for
    the guest collection creator.
    """
    assert g.sharing_policy is not None
    assert g.collection is not None and g.collection.is_mapped_collection

    sharing_policy_id = g.sharing_policy.id

    if sharing_policy_id == models.SharingPolicy.default_collection_policy(
        g.collection.id
    ):
        raise errors.BadRequestError(
            "The default sharing policy for a collection can not be deleted"
        )

    def delete_policy(collection):
        collection.delete_sharing_policy(sharing_policy_id)

    update_object(g.collection, delete_policy)

    res = api_schema.result_schema.dump(
        Result(message="Deleted sharing policy {}".format(str(sharing_policy_id)))
    )
    g.deleted_objects.append(sharing_policy_id)

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response
