from typing import List
import globus.manager.api.errors as errors
from globus.manager.constants import ConfigurationObjectTag
from .oidc_user_credential import OIDCUserCredential
from flask import current_app


class BoxUserCredential(OIDCUserCredential):
    POLICY_DATA_TYPE = "box_user_credential_policies#1.1.0"
    POLICY_DATA_TYPES = ["box_user_credential_policies#1.0.0", POLICY_DATA_TYPE]
    CONNECTOR_ID = "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63"

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.USER_CREDENTIAL.name + CONNECTOR_ID
        )
    }

    @property
    def connector_properties(self) -> List[str]:
        return super().connector_properties + ["max_upload"]

    def _credentials_to_patch(self, credentials):
        patch = super()._credentials_to_patch(credentials)
        if credentials:
            id_token = credentials.id_token
            if "policies" not in patch:
                patch["policies"] = {"DATA_TYPE": self.POLICY_DATA_TYPE}
            if "max_upload" in id_token:
                patch["policies"]["max_upload"] = id_token["max_upload"]
        return patch

    def __init__(
        self, appsync_data=None, doc=None, storage_gateway=None, credentials=None
    ):
        if doc:
            self.username = doc["username"]
            new_doc = {**doc}
            new_doc.pop("display_name", None)
            # Policies must be either empty (for an unprovisioned credential)
            # or passed in as a credentials object from the oauth callback
            policies = new_doc.get("policies", None)
            if policies is not None and policies not in [
                {"DATA_TYPE": data_type} for data_type in self.POLICY_DATA_TYPES
            ]:
                raise ValueError("policies")
            doc = new_doc

        super().__init__(
            appsync_data=appsync_data, doc=doc, storage_gateway=storage_gateway
        )
        if doc and self.uses_service_auth:
            token_info = self._get_enterprise_user_token()
            p = doc.get("policies", {})
            policies = {
                **p,
                **{
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                    "access_token": token_info["token"],
                    "token_expiry": token_info["token_expiry"],
                    "refresh_token": "enterprise",
                    "sub": token_info["sub"],
                    "email": token_info["email"],
                    "max_upload": token_info["max_upload"],
                },
            }
            doc = {
                **doc,
                **{
                    "display_name": token_info["name"],
                    "invalid": False,
                    "deleted": False,
                    "provisioned": True,
                    "policies": policies,
                },
            }
            self._patch(**doc)
        elif doc:
            patch = self._credentials_to_patch(credentials)
            patch["deleted"] = False
            patch["provisioned"] = bool(credentials)
            self._patch(**patch)

    max_upload = OIDCUserCredential.connector_property("max_upload")

    @property
    def connector_id(self):
        return self.CONNECTOR_ID

    @connector_id.setter
    def connector_id(self, value):
        pass

    @property
    def uses_service_auth(self) -> bool:
        return not self.storage_gateway.user_credential_required

    @property
    def supports_revoke(self) -> bool:
        return not self.uses_service_auth

    @property
    def token_uri(self) -> str:
        return "https://api.box.com/oauth2/token"

    def needs_refresh(self):
        if self.uses_service_auth and not getattr(self, "refresh_token", None):
            current_app.logger.debug("1.0.0 credential needs refresh")
            return True
        else:
            return super().needs_refresh()

    def refresh(self):
        if self.uses_service_auth:
            token_info = self._get_enterprise_user_token()
            self.refresh_token = "enterprise"
            self.access_token = token_info["token"]
            self.token_expiry = token_info["token_expiry"]
            self.sub = token_info["sub"]
            self.email = token_info["email"]
            self.max_upload = token_info["max_upload"]
            self.invalid = False
            self.provisioned = True
            self.configuration_hash.changed()
            current_app.logger.debug("Token refreshed: {}".format(self.token_expiry))
        else:
            super().refresh()

    def _get_enterprise_user_token(self):
        try:
            token_info = self.storage_gateway.get_user_access_token(self.username)
        except Exception as e:
            current_app.logger.warning("Enterprise user token failure: %s", str(e))
            raise errors.InvalidCredentialError(user_credential_id=self.id)
        return token_info

    def __repr__(self):
        return "<BoxUserCredential {}>".format(self.id)
