import ipaddress

import marshmallow as ma
from marshmallow_oneofschema import OneOfSchema

from globus.manager.constants import NodeStatus
from .data_type import DataTypeField, DataTypeSchema
from .schema_version_checker import version_safe_type_schemas
from ..enum_field import EnumField
from ..patch_schema import create_patch_schema


class NodeSchema_1_0_0(ma.Schema):
    __doc__ = """
    Services for Globus Connect Server endpoints may be deployed on multiple
    different physical resources, referred to as data transfer nodes. Each node
    may have one or more IP addresses, TCP incoming and outgoing port ranges,
    and a *status* value indicating whether it is configured to actively
    respond to requests or is in maintenance mode.
    """

    _DATA_TYPE = "node#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    id = ma.fields.UUID(
        description="""
        Unique id string this node. This is system generated and may not
        be included in create requests.
        """
    )
    ip_addresses = ma.fields.List(
        ma.fields.String,
        required=True,
        description="List of IP addresses for this node",
    )

    incoming_port_range = ma.fields.List(
        ma.fields.Int(validate=[ma.validate.Range(min=0, max=2**16 - 1)]),
        validate=[ma.validate.Length(equal=2)],
        required=False,
        description="Allowed port range for incoming TCP data connections",
    )

    outgoing_port_range = ma.fields.List(
        ma.fields.Int(validate=[ma.validate.Range(min=0, max=2**16 - 1)]),
        validate=[ma.validate.Length(equal=2)],
        required=False,
        description="Port range used as the source for outgoing TCP data connections",
    )

    status = EnumField(
        NodeStatus,
        required=True,
        description="""
        Current status of the Node. If a Node is marked *inactive*, it will
        be removed from the DNS entries for this endpoint and will return
        an error on any attempt to use the Manager API or attempt a
        Transfer using this node.
        """,
    )

    @ma.validates("ip_addresses")
    def validate_ip_addresses(self, data, **kwargs):
        if data:
            try:
                for i in data:
                    ipaddress.ip_address(i)
            except Exception as e:
                raise ma.ValidationError(str(e))


class NodeSchema_1_1_0(NodeSchema_1_0_0):
    __doc__ = (
        NodeSchema_1_0_0.__doc__
        + """
    Version 1.1.0 adds support for setting the data interface on a node.
    """
    )
    _DATA_TYPE = "node#1.1.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    data_interface = ma.fields.String(
        required=False,
        allow_none=True,
        description="IP address on which this node listens for data transfers",
    )

    @ma.validates("data_interface")
    def validate_data_interface(self, data, **kwargs):
        if data:
            try:
                ipaddress.ip_address(data)
            except Exception as e:
                raise ma.ValidationError(str(e))


node_1_0_0_schema = NodeSchema_1_0_0()
node_1_0_0_patch_schema = create_patch_schema(NodeSchema_1_0_0)

node_1_1_0_schema = NodeSchema_1_1_0()
node_1_1_0_patch_schema = create_patch_schema(NodeSchema_1_1_0)


class Node(OneOfSchema):
    __doc__ = NodeSchema_1_1_0.__doc__

    _DATA_TYPE = node_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = version_safe_type_schemas(
        {"node#1.0.0": node_1_0_0_schema, "node#1.1.0": node_1_1_0_schema}
    )
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE


class NodePatch(OneOfSchema):
    __doc__ = NodeSchema_1_1_0.__doc__
    _DATA_TYPE = node_1_1_0_patch_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = version_safe_type_schemas(
        {"node#1.0.0": node_1_0_0_patch_schema, "node#1.1.0": node_1_1_0_patch_schema}
    )
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE


node_schema = Node()
node_patch_schema = NodePatch(partial=True)

DataTypeSchema.register_schema(node_1_0_0_schema)
DataTypeSchema.register_schema(node_1_1_0_schema)
