from typing import Any, Dict, List, Optional, Mapping
from typing_extensions import Literal

import marshmallow as ma
from mypy_extensions import TypedDict

from ..api.domain import domain_private_schema, DomainDict
from ..api.schema_version_checker import verify_schema_version


OIDCClientDict = TypedDict(
    "OIDCClientDict",
    {
        "client_secret": str,
        "client_salt": Literal["NOT_ACTUALLY_USED"],
        "redirect_uris": List[List[Optional[str]]],
    },
)

OIDCServerDict = TypedDict(
    "OIDCServerDict",
    {
        "DATA_TYPE": Literal["oidc_server#1.0.0"],
        "client_id": Optional[str],
        "client_secret": Optional[str],
        "domain_name": Optional[str],
        "environment": Optional[str],
        "display_name": Optional[str],
        "support_contact": str,
        "support_email": str,
        "pam_service": str,
        "shared_key": str,
        "shared_session_key": str,
        "clients": Mapping[str, OIDCClientDict],
        "jwk": Mapping[str, Any],
        "domain": DomainDict,
    },
    total=False,
)


class OIDCClientSchema(ma.Schema):
    """
    Marshmallow schema for the values of the clients dict in the
    OIDCServerYamlSchema and the OIDCServerSchema
    """

    client_secret = ma.fields.String()
    client_salt = ma.fields.Constant("NOT_ACTUALLY_USED")
    redirect_uris = ma.fields.List(
        ma.fields.List(ma.fields.Url(schemes=["https"], allow_none=True))
    )


class NestingAwareSchema(ma.Schema):
    @ma.post_dump
    def nest_output(self, data, *, many: bool = False, **kwargs):
        newret = {}  # type: Dict[str, Any]
        for k, v in data.items():
            subkey = k
            subret = newret
            while "." in subkey:
                subdoc, subkey = subkey.split(".", 1)
                if subdoc not in subret:
                    subret[subdoc] = {}
                subret = subret[subdoc]
            subret[subkey] = v
        return newret

    @ma.pre_load
    def unnest_input(self, data, *, many: bool = False, **kwargs):
        newret = {}
        for k, v in self.fields.items():
            subkey = v.data_key if v.data_key else k
            subret = data
            while "." in subkey:
                subdoc, subkey = subkey.split(".", 1)
                if subdoc in subret:
                    subret = subret[subdoc]
                else:
                    subret = None
                    break
            if subret and subkey in subret:
                newret[v.data_key if v.data_key else k] = subret[subkey]
        return newret


class OIDCServerYamlSchema(NestingAwareSchema):
    """
    Marshmallow schema for the yaml file that globus-oidc consumes.
    """

    client_id = ma.fields.String(data_key="auth_client.client_id", allow_none=True)
    client_secret = ma.fields.String(
        data_key="auth_client.client_secret", allow_none=True
    )
    domain = ma.fields.String(
        data_key="auth_client.domain", attribute="domain.domain_name", allow_none=True
    )
    environment = ma.fields.String(data_key="auth_client.env", allow_none=True)
    display_name = ma.fields.String(
        data_key="oidc_server.display_name", allow_none=True
    )
    support_contact = ma.fields.String(data_key="oidc_server.support_contact")
    support_email = ma.fields.Email(data_key="oidc_server.support_email")
    pam_service = ma.fields.String(data_key="oidc_server.pam_service")
    shared_key = ma.fields.String(data_key="oidc_server.shared_key", allow_none=True)
    shared_session_key = ma.fields.String(
        data_key="oidc_server.shared_session_key", allow_none=True
    )
    clients = ma.fields.Mapping(
        keys=ma.fields.String,
        values=ma.fields.Nested(OIDCClientSchema),
    )
    jwk = ma.fields.Raw(data_key="oidc_server.jwk")

    private = ["client_secret", "shared_key", "shared_session_key", "jwk"]

    @ma.post_dump
    def remove_private_clients(self, data, **kwargs):
        if self.load_only and "clients" in data:
            for c in data["clients"].values():
                if "client_secret" in c:
                    del c["client_secret"]
        return data


class OIDCServerSchema(ma.Schema):
    """
    Marshmallow schema used by the GCS model. This is the data stored in appsync
    and shared between DTNs.
    """

    _DATA_TYPE = "oidc_server#1.0.0"

    DATA_TYPE = ma.fields.String(
        missing=_DATA_TYPE, allow_none=False, default=_DATA_TYPE
    )

    client_id = ma.fields.String(allow_none=True)
    client_secret = ma.fields.String(allow_none=True)
    environment = ma.fields.String(allow_none=True)

    display_name = ma.fields.String(allow_none=True)
    support_contact = ma.fields.String(required=True)
    support_email = ma.fields.Email(required=True)
    pam_service = ma.fields.String()
    shared_key = ma.fields.String(allow_none=True)
    shared_session_key = ma.fields.String(allow_none=True)

    jwk = ma.fields.Raw()

    clients = ma.fields.Mapping(
        keys=ma.fields.String,
        values=ma.fields.Nested(OIDCClientSchema),
    )

    domain = ma.fields.Nested(domain_private_schema)
    deleted = ma.fields.Boolean()

    @ma.validates("DATA_TYPE")
    def validate_data_type(self, data, **kwargs):
        verify_schema_version(data, self._DATA_TYPE)


oidc_server_schema = OIDCServerSchema()
oidc_server_yaml_private_schema = OIDCServerYamlSchema()
oidc_server_yaml_schema = OIDCServerYamlSchema(load_only=OIDCServerYamlSchema.private)
