from typing import List

import marshmallow as ma
from marshmallow_oneofschema import OneOfSchema

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeField, DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


BOX_CONNECTOR_ID = "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63"


class BoxAppAuth(ma.Schema):
    """
    Key information used to perform JWT grants for using the Box API
    """

    publicKeyID = ma.fields.Str(required=True, description="ID of the public key")
    privateKey = ma.fields.Str(required=True, description="Private key")
    passphrase = ma.fields.Str(
        required=True, description="Passphrase to decrypt the private key"
    )


class BoxAppSettings(ma.Schema):
    """
    Values from the Box JWT client configuration that the storage gateway uses
    to identify and authenticate with the Box API.  This is only set when
    configuring the storage gateway for Box enterprise authentication.
    """

    clientID = ma.fields.Str(required=True, description="Application client ID")
    clientSecret = ma.fields.Str(required=True, description="Application client secret")
    appAuth = ma.fields.Nested(
        BoxAppAuth, required=True, description="Box application keys"
    )


class BoxJsonConfig(ma.Schema):
    boxAppSettings = ma.fields.Nested(
        BoxAppSettings,
        required=True,
        description="Box Application settings",
    )
    enterpriseID = ma.fields.Str(
        required=True,
        description="""
        Identifies which Box Enterprise this storage gateway is authorized
        access to. This is only set when configuring the storage gateway
        for Box enterprise authentication.
        """,
    )
    private = ["boxAppSettings", "enterpriseID"]


class BoxStorage_1_0_0(BoxJsonConfig):
    """
    Connector-specific storage gateway policies for the Box connector.
    """

    _DATA_TYPE = "box_storage_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    connector_id = BOX_CONNECTOR_ID


class BoxStorage_1_1_0(ma.Schema):
    """
    Connector-specific storage gateway policies for the Box connector.
    """

    _DATA_TYPE = "box_storage_policies#1.1.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    boxAppSettings = ma.fields.Nested(BoxAppSettings, required=False)
    enterpriseID = ma.fields.Str(
        required=False,
        description="""
        Identifies which Box Enterprise this storage gateway is authorized
        access to. This is only set when configuring the storage gateway for
        Box enterprise authentication.
        """,
    )
    client_id = ma.fields.Str(
        required=False,
        description="""
        Client ID of the Box OAuth2 application registered on the Box developer
        console.  This is only set when configuring the storage gateway for
        OAuth2 user authentication.
        """,
    )
    secret = ma.fields.Str(
        required=False,
        description="""
        Secret associated with the client_id set in this policy.  This is only
        set when configuring the storage gateway for OAuth2 user
        authentication.
        """,
    )
    user_api_rate_limit = ma.fields.Int(
        required=False,
        description="""
        User API Rate Limit associated with this client ID in operations per
        second per user.
        """,
    )
    user_credential_required = ma.fields.Bool(
        dump_only=True,
        description="""
        Flag indicating whether users must register a credential.  If true,
        this storage gateway is configured for OAuth2 user authentication.  If
        false (and for older DATA_TYPE where this property is missing), this
        storage gateway is configured for enterprise authentication.
        """,
    )
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        description="""
        URL of the auth callback that must be set on the Box developer
        console for the Box application of client_id.
        """,
    )

    private = [
        "client_id",
        "secret",
        "user_api_rate_limit",
        "boxAppSettings",
        "enterpriseID",
    ]
    connector_id = BOX_CONNECTOR_ID

    @ma.pre_load
    def ignore_dump_only(self, data, **kwargs):
        """
        Ignore dump_only schema fields when loading.
        """
        for k, v in self.fields.items():
            if v.dump_only:
                data.pop(k, None)
        return data

    @ma.validates_schema
    def old_or_new(self, data, **kwargs):
        # Either (boxAppSettings, enterpriseID) or (client_id, secret)
        partial = kwargs.get("partial", False)
        if partial:
            old_style = "boxAppSettings" in data and "enterpriseID" in data
            new_style = "client_id" in data and "secret" in data
            if not old_style and not new_style:
                return
        else:
            old_style = "boxAppSettings" in data or "enterpriseID" in data
            new_style = "client_id" in data or "secret" in data

        if not (old_style ^ new_style):
            raise ma.ValidationError(
                f"The {self._DATA_TYPE} document must contain either "
                f"boxAppSettings and enterpriseID or client_id and secret {kwargs}"
            )


box_storage_policies_1_0_0_schema = BoxStorage_1_0_0(load_only=BoxStorage_1_0_0.private)
box_storage_policies_1_1_0_schema = BoxStorage_1_1_0(load_only=BoxStorage_1_1_0.private)
box_storage_policies_1_0_0_private_schema = BoxStorage_1_0_0()
box_storage_policies_1_1_0_private_schema = BoxStorage_1_1_0()


class BoxStoragePolicies(OneOfSchema):
    """
    Connector-specific storage gateway policies for the Box connector
    """

    _DATA_TYPE = box_storage_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "box_storage_policies#1.0.0": box_storage_policies_1_0_0_schema,
        "box_storage_policies#1.1.0": box_storage_policies_1_1_0_schema,
    }
    type_field_remove = False
    connector_id = BOX_CONNECTOR_ID

    def get_obj_type(self, obj):
        return self._DATA_TYPE

    private = box_storage_policies_1_1_0_schema.private


class BoxStoragePoliciesPrivate(OneOfSchema):
    _DATA_TYPE = box_storage_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "box_storage_policies#1.0.0": box_storage_policies_1_0_0_private_schema,
        "box_storage_policies#1.1.0": box_storage_policies_1_1_0_private_schema,
    }
    type_field_remove = False
    connector_id = BOX_CONNECTOR_ID

    def get_obj_type(self, obj):
        return self._DATA_TYPE


class BoxCollectionPolicies(ma.Schema):
    """
    Connector-specific collection policies for the Box connector
    """

    _DATA_TYPE = "box_collection_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    private = []  # type: List[str]
    connector_id = BOX_CONNECTOR_ID


class BoxUserCredential_1_0_0(ma.Schema):
    """
    Connector-specific user credential policies for the Box connector
    """

    _DATA_TYPE = "box_user_credential_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    private = []  # type: List[str]
    connector_id = BOX_CONNECTOR_ID


class BoxUserCredential_1_1_0(ma.Schema):
    """
    Connector-specific user credential policies for the Box connector
    """

    _DATA_TYPE = "box_user_credential_policies#1.1.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    sub = ma.fields.Str(allow_none=True, description="OAuth subject identifier claim")
    email = ma.fields.Email(allow_none=True, description="OAuth email identifier claim")
    access_token = ma.fields.Str(
        required=False, allow_none=True, description="OAuth access token"
    )
    refresh_token = ma.fields.Str(
        required=False, allow_none=True, description="OAuth refresh token"
    )
    token_expiry = ma.fields.AwareDateTime(
        description="OAuth access token expiration time"
    )
    scopes = ma.fields.List(
        ma.fields.Str,
        description="OAuth scopes associated with this access token",
        example=["box_readwrite"],
    )
    max_upload = ma.fields.Int()

    private = ["access_token", "refresh_token", "max_upload"]
    connector_id = BOX_CONNECTOR_ID


box_storage_policies_private_schema = BoxStoragePoliciesPrivate()
box_storage_policies_schema = BoxStoragePolicies(load_only=BoxStoragePolicies.private)


box_user_credential_policies_1_0_0_schema = BoxUserCredential_1_0_0(
    load_only=BoxUserCredential_1_0_0.private
)
box_user_credential_policies_1_1_0_schema = BoxUserCredential_1_1_0(
    load_only=BoxUserCredential_1_1_0.private
)
box_user_credential_policies_1_0_0_private_schema = BoxUserCredential_1_0_0()
box_user_credential_policies_1_1_0_private_schema = BoxUserCredential_1_1_0()
box_collection_policies_schema = BoxCollectionPolicies(
    load_only=BoxCollectionPolicies.private
)
box_collection_policies_private_schema = BoxCollectionPolicies()


class BoxUserCredentialPolicies(OneOfSchema):
    _DATA_TYPE = box_user_credential_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "box_user_credential_policies#1.0.0": box_user_credential_policies_1_0_0_schema,
        "box_user_credential_policies#1.1.0": box_user_credential_policies_1_1_0_schema,
    }
    type_field_remove = False
    connector_id = BOX_CONNECTOR_ID

    def get_obj_type(self, obj):
        return self._DATA_TYPE

    private = box_user_credential_policies_1_1_0_schema.private + []


class BoxUserCredentialPoliciesPrivate(OneOfSchema):
    _DATA_TYPE = box_user_credential_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "box_user_credential_policies#1.0.0": box_user_credential_policies_1_0_0_private_schema,
        "box_user_credential_policies#1.1.0": box_user_credential_policies_1_1_0_private_schema,
    }
    type_field_remove = False
    connector_id = BOX_CONNECTOR_ID

    def get_obj_type(self, obj):
        return self._DATA_TYPE


box_user_credential_policies_schema = BoxUserCredentialPolicies(
    load_only=BoxUserCredentialPolicies.private
)
box_user_credential_policies_private_schema = BoxUserCredentialPoliciesPrivate()


DataTypeSchema.register_schema(box_storage_policies_1_0_0_schema)
DataTypeSchema.register_schema(box_storage_policies_1_1_0_schema)
DataTypeSchema.register_schema(box_user_credential_policies_1_0_0_schema)
DataTypeSchema.register_schema(box_user_credential_policies_1_1_0_schema)
DataTypeSchema.register_schema(box_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=box_storage_policies_schema, private=box_storage_policies_private_schema
)

UserCredentialSchema.register_schema(
    public=box_user_credential_policies_schema,
    private=box_user_credential_policies_private_schema,
)

CollectionSchema.register_schema(
    public=box_collection_policies_schema,
    private=box_collection_policies_private_schema,
)
