from typing import List

import marshmallow as ma

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeField, DataTypeSchema
from ..api.user_credential import UserCredentialSchema
from ..api.storage_gateway import StorageGatewaySchema


CEPH_CONNECTOR_ID = "1b6374b0-f6a4-4cf7-a26f-f262d9c6ca72"


class CephStoragePolicies(ma.Schema):
    """
    Connector-specific storage gateway policies for the Ceph connector
    """

    _DATA_TYPE = "ceph_storage_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    s3_endpoint = ma.fields.Url(description="URL of the Ceph RADOS Gateway S3 API")
    s3_buckets = ma.fields.List(
        ma.fields.Str,
        description="""
        List of buckets not owned by the collection owner that will be shown in
        the root of collections created at the base of this Storage Gateway.
        """,
    )
    ceph_admin_key_id = ma.fields.Str(
        description="""
        Administrator key id used to authenticate with the ceph admin service
        to obtain user credentials.
        """
    )
    ceph_admin_secret_key = ma.fields.Str(
        description="""
        Administrator secret key used to authenticate with the ceph admin
        service to obtain user credentials.
        """
    )

    private = ["ceph_admin_key_id", "ceph_admin_secret_key"]
    connector_id = CEPH_CONNECTOR_ID


class CephCollectionPolicies(ma.Schema):
    """
    Connector-specific collection policies for the Ceph connector
    """

    _DATA_TYPE = "ceph_collection_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    private = []  # type: List[str]
    connector_id = CEPH_CONNECTOR_ID


class CephUserCredentialPolicies(ma.Schema):
    """
    Connector-specific user credential policies for the Ceph connector
    """

    _DATA_TYPE = "ceph_user_credential_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    private = []  # type: List[str]
    connector_id = CEPH_CONNECTOR_ID


ceph_storage_policies_private_schema = CephStoragePolicies()
ceph_storage_policies_schema = CephStoragePolicies(
    load_only=CephStoragePolicies.private
)

ceph_user_credential_policies_schema = CephUserCredentialPolicies(
    load_only=CephUserCredentialPolicies.private
)
ceph_user_credential_policies_private_schema = CephUserCredentialPolicies()
ceph_collection_policies_schema = CephCollectionPolicies(
    load_only=CephCollectionPolicies.private
)
ceph_collection_policies_private_schema = CephCollectionPolicies()

DataTypeSchema.register_schema(ceph_storage_policies_schema)
DataTypeSchema.register_schema(ceph_user_credential_policies_schema)
DataTypeSchema.register_schema(ceph_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=ceph_storage_policies_schema, private=ceph_storage_policies_private_schema
)

UserCredentialSchema.register_schema(
    public=ceph_user_credential_policies_schema,
    private=ceph_user_credential_policies_private_schema,
)

CollectionSchema.register_schema(
    public=ceph_collection_policies_schema,
    private=ceph_collection_policies_private_schema,
)
