from typing import List

import marshmallow as ma

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeField, DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


ONEDRIVE_CONNECTOR_ID = "28ef55da-1f97-11eb-bdfd-12704e0d6a4d"


class OneDriveStoragePolicies(ma.Schema):
    """
    Connector-specific storage gateway policies for the OneDrive connector
    """

    _DATA_TYPE = "onedrive_storage_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    client_id = ma.fields.Str(
        required=True,
        allow_none=True,
        description="Client ID registered with the Azure console to access OneDrive",
    )
    secret = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        Secret created in the Azure console to access OneDrive with the
        client_id in this policy.
        """,
    )
    tenant = ma.fields.Str(
        required=False,
        allow_none=True,
        description="""
        Tenant ID of the Microsoft organization. Required when Supported
        Account Types of the Azure application is set to Single tenant.
        """,
    )
    user_api_rate_limit = ma.fields.Int(
        required=False,
        description="""
        User API Rate Limit associated with this client ID in operations per
        second per user.
        """,
    )
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        description="""
        URL of the auth callback that must be registered on the Microsoft
        API console for the application client_id in order to process
        Microsoft credentials.
        """,
    )

    private = ["client_id", "secret", "tenant", "user_api_rate_limit"]
    connector_id = ONEDRIVE_CONNECTOR_ID

    @ma.pre_load
    def ignore_dump_only(self, data, **kwargs):
        """
        Ignore dump_only schema fields when loading.
        """
        for k, v in self.fields.items():
            if v.dump_only:
                data.pop(k, None)
        return data


class OneDriveCollectionPolicies(ma.Schema):
    """
    Connector-specific collection policies for the OneDrive connector
    """

    _DATA_TYPE = "onedrive_collection_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    private = []  # type: List[str]
    connector_id = ONEDRIVE_CONNECTOR_ID


class OneDriveUserCredentialPolicies(ma.Schema):
    """
    Connector-specific user credential policies for the OneDrive connector
    """

    _DATA_TYPE = "onedrive_user_credential_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    sub = ma.fields.Str(allow_none=True, description="OAuth subject identifier claim")
    email = ma.fields.Email(allow_none=True, description="OAuth email claim")
    access_token = ma.fields.Str(
        required=False,
        allow_none=True,
        description="OAuth access token",
    )

    refresh_token = ma.fields.Str(
        required=False,
        allow_none=True,
        description="OAuth refresh token",
    )
    token_expiry = ma.fields.AwareDateTime(
        description="OAuth access token expiration time"
    )
    scopes = ma.fields.List(
        ma.fields.Str,
        description="OAuth scopes associated with the access token",
        example=["openid", "email", "profile", "offline_access", "files.readwrite.all"],
    )

    tid = ma.fields.Str()

    private = ["access_token", "refresh_token"]
    connector_id = ONEDRIVE_CONNECTOR_ID


onedrive_user_credential_policies_private_schema = OneDriveUserCredentialPolicies()
onedrive_user_credential_policies_schema = OneDriveUserCredentialPolicies(
    load_only=OneDriveUserCredentialPolicies.private
)
onedrive_storage_policies_private_schema = OneDriveStoragePolicies()
onedrive_storage_policies_schema = OneDriveStoragePolicies(
    load_only=OneDriveStoragePolicies.private
)
onedrive_collection_policies_schema = OneDriveCollectionPolicies(
    load_only=OneDriveCollectionPolicies.private
)
onedrive_collection_policies_private_schema = OneDriveCollectionPolicies()

DataTypeSchema.register_schema(onedrive_storage_policies_schema)
DataTypeSchema.register_schema(onedrive_user_credential_policies_schema)
DataTypeSchema.register_schema(onedrive_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=onedrive_storage_policies_schema,
    private=onedrive_storage_policies_private_schema,
)

UserCredentialSchema.register_schema(
    public=onedrive_user_credential_policies_schema,
    private=onedrive_user_credential_policies_private_schema,
)

CollectionSchema.register_schema(
    public=onedrive_collection_policies_schema,
    private=onedrive_collection_policies_private_schema,
)
