import datetime
import json
import os
import sqlite3
import textwrap
from typing import Any, Dict, List, Optional, Union
from mypy_extensions import TypedDict
import uuid

from flask import current_app
import globus_sdk

from .acme_account import AcmeAccountV53
from .collection import CollectionV53, GoogleCloudStorageCollectionV53
from .endpoint import EndpointV53
from .storage_gateway import StorageGatewayV53
from .user_credential import UserCredentialV53

import globus.manager.api.models as models
from globus.manager.constants import CollectionType


UserCredentialRow = TypedDict(
    "UserCredentialRow",
    {
        "created_at": str,
        "updated_at": str,
        "id": str,
        "display_name": str,
        "identity_id": str,
        "storage_credential": Optional[Union[str, Dict[str, Any]]],
        "storage_gateway_id": str,
        "storage_gateway_user_id": str,
        "deleted": int,
    },
)


class ConfigurationDataV53:
    """
    Configuration state from a GCSv5.3 configuration database. On initialization,
    this class parses the state from the gcsv5.3 database and creates the V53
    compatibility models which are used to migrate the configuration to a v5.4
    database.
    """

    UNDETERMINED_UUID_PREFIX = "00000000-0000-0000-0000-"

    def __init__(self, owner):
        self._auth_client = None
        self._transfer_client = None

        with sqlite3.connect(os.path.expanduser("~gcsweb/gcs.db")) as con:
            con.row_factory = sqlite3.Row
            cursor = con.cursor()

            # key is the configuration key
            self.configuration = {}  # type: Dict[str, str]
            # key is the storage gateway id
            self.storage_gateway = {}  # type: Dict[str, models.StorageGateway]
            # key is the user credential id
            self.user_credential = {}  # type: Dict[str, models.UserCredential]
            self.acme_account = None  # type: Optional[models.AcmeAccount]

            self.domain = None  # type: models.GlobusDnsDomain

            # Key is the collection id. For mapped collections created as part
            # of the cleanup phase, the collection_id starts with
            # UNDETERMINED_UUID_PREFIX
            self.collection = {}  # type: Dict[str, CollectionV53]

            # Key is the role id
            self.role = {}  # type: Dict[str, models.Role]

            # Key is the storage_gateway id
            self.allow_guest_collections = {}  # type: Dict[str, bool]
            self.allow_mapped_collections = {}  # type: Dict[str, bool]
            self.sharing_users_allow = {}  # type: Dict[str, bool]
            self.sharing_users_deny = {}  # type: Dict[str, bool]
            self.sharing_groups_allow = {}  # type: Dict[str, bool]
            self.sharing_groups_deny = {}  # type: Dict[str, bool]

            # Map credential id (which may be for a dummy credential to a
            # StorageGateway model
            self.storage_gateway_for_credential_id = (
                {}
            )  # type: Dict[str, models.StorageGateway]
            self.owner_id = None  # type: str
            # Key is the storage_gateway id, value is the root
            self.storage_gateway_root = {}  # Dict[str, str]

            self.endpoint = None  # type: models.Endpoint

            # Initialize data from database and configuration files
            self._init_configuration_table(cursor)

            self._check_endpoint_id()
            self._init_owner(owner)
            self._init_storage_gateway_table(cursor)
            self._init_user_credential_table(cursor)
            self._init_collection_table(cursor)
            self._init_roles()
            self._init_domain()
            self._init_letsencrypt_account()
            self._init_endpoint()

        # Some post-processing of the migration models to adjust for
        # differences
        self._cleanup_collections()

    def _init_configuration_table(self, c: sqlite3.Cursor):
        """
        Load the configuration table into self.configuration
        """
        for row in c.execute(
            "select created_at, updated_at, name, value from configuration"
        ):
            if row["name"] == "environment":
                value = row["value"].lower()
                if value == "default":
                    value = "production"
            else:
                value = row["value"]
            self.configuration[row["name"]] = value
            current_app.config[row["name"].upper()] = value

    def _check_endpoint_id(self):
        host_endpoint_id = self.configuration["host_endpoint_id"]
        client_id = self.configuration["client_id"]
        try:
            transfer_doc = self.transfer_client.get_endpoint(host_endpoint_id).data
        except globus_sdk.exc.GlobusAPIError as e:
            if e.http_status == 404:
                transfer_doc = self.transfer_client.get_endpoint(client_id).data
                if transfer_doc:
                    host_endpoint_id = client_id
                    self.configuration["host_endpoint_id"] = client_id
            else:
                raise

    def _init_owner(self, owner: str) -> None:
        owner_identities = self.auth_client.get_identities(usernames=[owner]).data.get(
            "identities", []
        )
        if len(owner_identities) != 1:
            raise Exception(
                "\n".join(
                    textwrap.wrap(
                        "Unable to verify that Globus user {owner} exists. If "
                        "Google is used by this account to log into Globus, "
                        "try using the account name "
                        "{owner}@accounts.google.com. Otherwise, make sure you "
                        "have the correct Globus account name.".format(owner=owner)
                    )
                )
            )

        owner_identity = owner_identities[0]
        if owner_identity["status"] not in ["used", "private"]:
            raise Exception(
                "\n".join(
                    textwrap.wrap(
                        "The requested owner account exists as a Globus "
                        "identity, but it has not been used yet."
                    )
                    + textwrap.wrap(
                        "If that account uses Google to log into Globus, try "
                        "using {}@accounts.google.com as the owner. Otherwise, "
                        "make sure that the account has logged in to globus.org "
                        "before assigning it as an endpoint owner.".format(owner)
                    )
                )
            )
        self.owner_id = owner_identity["id"]

    def _init_storage_gateway_table(self, c: sqlite3.Cursor):
        """
        Load the storage_gateway table into self.storage_gateway. Values are
        stored in that method
        in that
        """
        for row in c.execute(
            "SELECT created_at, updated_at, id, display_name, credential, "
            "storage_policies, deleted, connector_id, allow_guest_collections, "
            "allow_mapped_collections, require_high_assurance "
            "FROM storage_gateway"
        ):
            parsed_row = {
                k: row[k]
                for k in row.keys()
                if k not in ["credential", "storage_policies"]
            }
            if row["credential"]:
                parsed_row["credential"] = json.loads(row["credential"])
            else:
                parsed_row["credential"] = None

            if row["storage_policies"]:
                parsed_row["storage_policies"] = json.loads(row["storage_policies"])
            else:
                parsed_row["storage_policies"] = None

            self.storage_gateway[row["id"]] = StorageGatewayV53(**parsed_row)
            self.allow_guest_collections[row["id"]] = row["allow_guest_collections"]
            self.allow_mapped_collections[row["id"]] = row["allow_mapped_collections"]

            # We need to save off these 5.3 POSIX storage gateway settings and later place them
            # on mapped collections attached to this gateway.
            private_policies = parsed_row.get("storage_policies", {}).get("private", {})
            if "users_allow" in private_policies:
                self.sharing_users_allow[row["id"]] = private_policies[
                    "users_allow"
                ].split(",")
            if "users_deny" in private_policies:
                self.sharing_users_deny[row["id"]] = private_policies[
                    "users_deny"
                ].split(",")
            if "groups_allow" in private_policies:
                self.sharing_groups_allow[row["id"]] = private_policies[
                    "groups_allow"
                ].split(",")
            if "groups_deny" in private_policies:
                self.sharing_groups_deny[row["id"]] = private_policies[
                    "groups_deny"
                ].split(",")

            self.storage_gateway_root[row["id"]] = parsed_row["storage_policies"][
                "public"
            ]["root"]

    def _init_user_credential_table(self, c: sqlite3.Cursor):
        """
        Load the user_credential table into self.user_credential.
        """
        for row in c.execute(
            "SELECT created_at, updated_at, id, display_name, identity_id, "
            "storage_credential, storage_gateway_id, storage_gateway_user_id, "
            "deleted FROM user_credential"
        ):
            storage_credential_str = row["storage_credential"]
            parsed_row = {
                "created_at": row["created_at"],
                "updated_at": row["updated_at"],
                "id": row["id"],
                "display_name": row["display_name"],
                "identity_id": row["identity_id"],
                "storage_credential": (
                    json.loads(storage_credential_str)
                    if storage_credential_str
                    else None
                ),
                "storage_gateway_id": row["storage_gateway_id"],
                "storage_gateway_user_id": row["storage_gateway_user_id"],
                "deleted": row["deleted"],
            }  # type: UserCredentialRow

            storage_gateway = self.storage_gateway[row["storage_gateway_id"]]
            self.storage_gateway_for_credential_id[row["id"]] = self.storage_gateway[
                row["storage_gateway_id"]
            ]
            if parsed_row["identity_id"] != ":globus-mapping:":
                self.user_credential[row["id"]] = UserCredentialV53(
                    storage_gateway, self.auth_client, **parsed_row
                )

    def _init_collection_table(self, c: sqlite3.Cursor):
        """
        Load the collection table into self.collection
        """
        transfer_client = self.transfer_client
        for row in c.execute(
            "SELECT created_at, updated_at, id, path, user_credential_id, domain, "
            "collection_policies FROM collection"
        ):
            uc = self.user_credential.get(row["user_credential_id"])
            sg = self.storage_gateway_for_credential_id[row["user_credential_id"]]
            connector_id = sg.connector_id
            try:
                transfer_doc = transfer_client.get_endpoint(row["id"]).data
            except globus_sdk.TransferAPIError as e:
                if e.http_status == 404:
                    # Skip deleted collection
                    continue
                else:
                    raise

            parsed_row = {k: row[k] for k in row.keys() if k != "collection_policies"}

            parsed_row["collection_policies"] = json.loads(row["collection_policies"])

            identity_mapping = self._parse_gridmap(parsed_row)

            self.collection[row["id"]] = CollectionV53(
                connector_id,
                transfer_doc,
                self.allow_guest_collections[sg.id],
                str(uc.identity_id) if uc else self.owner_id,
                uc,
                sg,
                self.storage_gateway_root[sg.id],
                identity_mapping,
                **parsed_row
            )

    def _make_role(self, resource=None, **kwargs):
        if resource:
            role = "collection:{}".format(kwargs["role"])
        else:
            role = "endpoint:{}".format(kwargs["role"])

        return models.Role(
            appsync_data={
                "configuration_hash": {
                    "DATA_TYPE": "role_config#1.0.0",
                    "id": kwargs["id"],
                    "principal_type": kwargs["principal_type"],
                    "identity_id": kwargs["principal"],
                    "resource": resource,
                    "role": role,
                }
            }
        )

    def _init_roles(self):
        """
        Load the roles from transfer for the endpoint and all collections
        """
        transfer_client = self.transfer_client
        host_endpoint_id = self.configuration["host_endpoint_id"]

        for role in transfer_client.endpoint_role_list(host_endpoint_id).data["DATA"]:
            self.role[role["id"]] = self._make_role(resource=None, **role)

        for col in self.collection.values():
            for role in transfer_client.endpoint_role_list(col.id).data["DATA"]:
                if role["principal"] == col.identity_id:
                    primary_role = True
                else:
                    primary_role = False
                self.role[role["id"]] = self._make_role(
                    resource=col.id, primary_role=primary_role, **role
                )

    def _init_domain(self):
        self.domain = models.GlobusDnsDomain(
            appsync_data={
                "configuration_hash": {
                    "DATA_TYPE": "endpoint_domain#1.0.0",
                    "domain_name": self.configuration["server_name"],
                }
            }
        )

    def _cleanup_collections(self):
        self._cleanup_mapped_collection_gridmaps()
        self._cleanup_mapped_collection_dummy_creds_and_identity_id()
        self._cleanup_guest_collection_mapped_collection_ids()

    def _cleanup_mapped_collection_gridmaps(self):
        """
        Cleanup the model change from v5.3 to v5.4.
        In v5.3, any mapped collection can have a gridmap file associated
        with it. If there are any mapped collections that have non-empty
        gridmap files, we'll clone the storage gateway and credentials and
        associate the identity_mapping generated from the gridmap
        """

        associated_collections = {
            storage_gateway_id: [
                collection
                for collection in self.collection.values()
                if collection.storage_gateway_id == storage_gateway_id
            ]
            for storage_gateway_id in self.storage_gateway.keys()
        }

        need_repair = {}
        for c in self.collection.values():
            if c.collection_type == CollectionType.mapped and c._identity_mapping:
                if len(associated_collections(c.storage_gateway_id)) > 1:
                    if c.storage_gateway_id not in need_repair:
                        need_repair[c.storage_gateway_id] = [c]
                    else:
                        need_repair[c.storage_gateway_id].append(c)

        for sg, collections in need_repair.items():
            for c in collections:
                new_sg = StorageGatewayV53(
                    **{**sg._kwargs, **{"id": str(uuid.uuid4())}}
                )
                new_sg.display_name = "{} for collection {}".format(
                    new_sg.display_name, c.display_name
                )
                new_sg.identity_mappings = c._identity_mappings
                uc = self.user_credential[c.user_credential_id]
                new_uc = UserCredentialV53(**uc._kwargs)
                new_uc.storage_gateway_id = new_sg.id
                new_uc.id = str(uuid.uuid4())
                c.storage_gateway_id = new_sg.id
                c.user_credential_id = new_uc.id
                self.storage_gateway[new_sg.id] = new_sg
                self.user_credential[new_uc.id] = new_uc

    def _cleanup_mapped_collection_dummy_creds_and_identity_id(self):
        """
        In v5.3, mapped collections have a dummy credential but no identity_id
        indicating who created the mapped collection. In v5.4, every collection
        has a valid identity_id of the creator of the collection. Since the info
        is unavailable, we'll substitute the endpoint's client_id
        """
        for c in self.collection.values():
            if c.collection_type == CollectionType.mapped:
                c.identity_id = self.configuration["client_id"]
                if hasattr(self, "user_credential_id"):
                    delattr(self, "user_credential_id")

    def _cleanup_guest_collection_mapped_collection_ids(self):
        """
        In v5.3, Guest collections are created directly from storage gateways,
        but in v5.4, there must be a mapped collection for each storage gateway
        to be created, and there is a reference to the mapped collection.
        """
        new_id = 0
        new_id_format = "00000000-0000-0000-0000-{0:012d}"

        # Key is the storage gateway ID, value is the CollectionV53 object
        new_collections = {}
        now = str(datetime.datetime.now(tz=datetime.timezone.utc))
        for c in self.collection.values():
            if c.collection_type == CollectionType.guest:
                if c.storage_gateway_id not in new_collections:
                    sg = self.storage_gateway[c.storage_gateway_id]

                    collection_policies = {
                        "public": {"type": "mapped", "allow_guest_collections": True},
                        "private": {},
                    }

                    if isinstance(c, GoogleCloudStorageCollectionV53):
                        project = c.project
                        collection_policies["private"][
                            "google_cloud_storage_project"
                        ] = project

                    new_uuid = new_id_format.format(new_id)
                    new_mc = CollectionV53(
                        c.connector_id,
                        {
                            "display_name": "Migration {} {}".format(
                                self.endpoint.endpoint_id,
                                new_id,
                            ),
                            "default_directory": "/",
                            "public": True,
                            "force_encryption": c.force_encryption,
                            "disable_verify": c.disable_verify,
                            "keywords": "migration:{}".format(sg.id),
                        },
                        True,
                        self.owner_id,
                        None,
                        sg,
                        self.storage_gateway_root[sg.id],
                        None,
                        id=new_uuid,
                        path="/",
                        created_at=now,
                        updated_at=now,
                        collection_policies=collection_policies,
                    )
                    new_collections[c.storage_gateway_id] = new_mc
                    c.mapped_collection_id = new_mc.id
                    new_id += 1
                else:
                    c.mapped_collection_id = new_collections[c.storage_gateway_id].id

        self.collection.update({nc.id: nc for nc in new_collections.values()})

    def _parse_gridmap(self, row: Dict[str, Any]) -> Optional[List[Dict[str, Any]]]:
        """
        If a collection has a gridmap file with any entries, translate them into
        identity mapping expressions.
        """
        gridmap = row["collection_policies"].get("private", {}).get("gridmap")
        identity_mapping = []
        if gridmap:
            with open(gridmap, "r") as f:
                for line in f:
                    mapped_from, mapped_to = line.rsplit(sep=None, maxsplit=1)
                    for mapped_id in mapped_from.split(","):
                        mapped_id.replace('"', "")
                        identity_mapping.append(
                            {
                                "DATA_TYPE": "expression_identity_mapping#1.0.0",
                                "source": "{id}",
                                "match": mapped_id,
                                "literal": True,
                                "output": mapped_to,
                            }
                        )
        return identity_mapping if identity_mapping else None

    def _init_letsencrypt_account(self):
        """
        Parse the letsencrypt ACME account from the etc dir
        """
        letsencryptdir = os.path.join(
            os.path.expanduser(
                "~gcsweb/etc/letsencrypt/accounts/"
                "acme-v02.api.letsencrypt.org/directory"
            )
        )
        for acct in os.listdir(letsencryptdir):
            data = {}  # type: Dict[str, Any]
            for bn in ["meta", "private_key", "regr"]:
                fn = os.path.join(letsencryptdir, acct, bn + ".json")
                with open(fn, "r") as f:
                    data[bn] = json.load(f)
            assert self.acme_account is None
            self.acme_account = AcmeAccountV53(**data)

    def _init_endpoint(self):
        host_endpoint_id = self.configuration["host_endpoint_id"]
        client_id = self.configuration["client_id"]
        assert host_endpoint_id is not None

        transfer_doc = self.transfer_client.get_endpoint(host_endpoint_id).data

        transfer_doc.pop("host_endpoint_id", None)
        self.endpoint = EndpointV53(client_id, host_endpoint_id, **transfer_doc)
        data = transfer_doc.get("DATA", [])
        if data and len(data) > 0:
            server = data[0]
            incoming_start = server.get("incoming_data_port_start", 50000)
            incoming_end = server.get("incoming_data_port_end", 51000)
            outgoing_start = server.get("outgoing_data_port_start", None)
            outgoing_end = server.get("outgoing_data_port_end", None)

            if incoming_start and incoming_end:
                self.incoming_port_range = [incoming_start, incoming_end]
            else:
                self.incoming_port_range = None
            if outgoing_start and outgoing_end:
                self.outgoing_port_range = [outgoing_start, outgoing_end]
            else:
                self.outgoing_port_range = None

    @property
    def auth_client(self) -> globus_sdk.ConfidentialAppAuthClient:
        if not self._auth_client:
            client_id = self.configuration["client_id"]
            client_secret = self.configuration["client_secret"]
            environment = self.configuration.get("environment", "production").lower()

            if environment == "default":
                environment = "production"

            self._auth_client = globus_sdk.ConfidentialAppAuthClient(
                client_id, client_secret, environment=environment
            )
        return self._auth_client

    @property
    def transfer_client(self) -> globus_sdk.TransferClient:
        if not self._transfer_client:
            token = self.auth_client.oauth2_client_credentials_tokens()
            transfer_at = token.by_resource_server["transfer.api.globus.org"][
                "access_token"
            ]
            environment = self.configuration.get("environment", "production").lower()

            if environment == "default":
                environment = "production"

            self._transfer_client = globus_sdk.TransferClient(
                authorizer=globus_sdk.AccessTokenAuthorizer(transfer_at),
                environment=environment,
            )
        return self._transfer_client
